/**-----------------------------------------------------------------------------
 * \brief ̸
 * : HeightMap.cpp
 *
 * : HeightMap + Frustum Culling
 *       
 *------------------------------------------------------------------------------
 */

#define MAINBODY

#include <d3d9.h>
#include <d3dx9.h>
#include "ZCamera.h"
#include "ZFrustum.h"
#include "ZFLog.h"

#define WINDOW_W		500
#define WINDOW_H		500
#define WINDOW_TITLE	"HeightMap+Camera+Frustum"
#define BMP_HEIGHTMAP	"map128.bmp"

/**-----------------------------------------------------------------------------
 *  
 *------------------------------------------------------------------------------
 */
ZCamera*				g_pCamera = NULL;	// Camera Ŭ
ZFrustum*				g_pFrustum = NULL;	// Frustum Ŭ
HWND					g_hwnd = NULL;

LPDIRECT3D9             g_pD3D       = NULL; // D3D ̽  D3Dü
LPDIRECT3DDEVICE9       g_pd3dDevice = NULL; //   D3D̽
LPDIRECT3DVERTEXBUFFER9 g_pVB        = NULL; //   
LPDIRECT3DINDEXBUFFER9	g_pIB        = NULL; // ε  ε

LPDIRECT3DTEXTURE9		g_pTexHeight = NULL; // Texture ̸
LPDIRECT3DTEXTURE9		g_pTexDiffuse= NULL; // Texture 
D3DXMATRIXA16			g_matAni;
D3DXMATRIXA16			g_matWorld;
D3DXMATRIXA16			g_matView;
D3DXMATRIXA16			g_matProj;

DWORD					g_cxHeight = 0;			// ؽó ũ
DWORD					g_czHeight = 0;			// ؽó ũ
DWORD					g_dwMouseX = 0;			// 콺 ǥ
DWORD					g_dwMouseY = 0;			// 콺 ǥ
int						g_nTriangles = 0;		// ׷ ﰢ 
D3DXVECTOR3*			g_pvHeightMap = NULL;	// ̸  迭
BOOL					g_bHideFrustum = TRUE;	// Frustum ȱ׸ ΰ?
BOOL					g_bLockFrustum = FALSE;	// Frustum  ΰ?
BOOL					g_bWireframe = FALSE;	// ̾ ׸ΰ?

///    ü
struct CUSTOMVERTEX
{
	D3DXVECTOR3		p;
	D3DXVECTOR3		n;
	D3DXVECTOR2		t;
};

///   ü   Ÿ FVF
#define D3DFVF_CUSTOMVERTEX (D3DFVF_XYZ|D3DFVF_NORMAL|D3DFVF_TEX1)

struct MYINDEX
{
	WORD	_0, _1, _2;		/// WORD, 16Ʈ ε
};

/**-----------------------------------------------------------------------------
 * Direct3D ʱȭ
 *------------------------------------------------------------------------------
 */
HRESULT InitD3D( HWND hWnd )
{
    // ̽ ϱ D3Dü 
    if( NULL == ( g_pD3D = Direct3DCreate9( D3D_SDK_VERSION ) ) )
        return E_FAIL;

    // ̽  ü
    //  Ʈ ׸̱⶧, ̹ Z۰ ʿϴ.
    D3DPRESENT_PARAMETERS d3dpp;
    ZeroMemory( &d3dpp, sizeof(d3dpp) );
    d3dpp.Windowed = TRUE;
    d3dpp.SwapEffect = D3DSWAPEFFECT_DISCARD;
    d3dpp.BackBufferFormat = D3DFMT_UNKNOWN;
    d3dpp.EnableAutoDepthStencil = TRUE;
    d3dpp.AutoDepthStencilFormat = D3DFMT_D16;

    /// ̽ 
    if( FAILED( g_pD3D->CreateDevice( D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, hWnd,
                                      D3DCREATE_SOFTWARE_VERTEXPROCESSING,
                                      &d3dpp, &g_pd3dDevice ) ) )
    {
        return E_FAIL;
    }

    // ⺻ø, CCW
    g_pd3dDevice->SetRenderState( D3DRS_CULLMODE, D3DCULL_CCW );

    // Z۱ Ҵ.
    g_pd3dDevice->SetRenderState( D3DRS_ZENABLE, TRUE );

    return S_OK;
}

/**-----------------------------------------------------------------------------
 *  
 *------------------------------------------------------------------------------
 */
void InitMatrix()
{
	///   
	D3DXMatrixIdentity( &g_matWorld );
    g_pd3dDevice->SetTransform( D3DTS_WORLD, &g_matWorld );

    ///   
    D3DXVECTOR3 vEyePt( 0.0f, 100.0f, -(float)g_czHeight );
    D3DXVECTOR3 vLookatPt( 0.0f, 0.0f, 0.0f );
    D3DXVECTOR3 vUpVec( 0.0f, 1.0f, 0.0f );
    D3DXMatrixLookAtLH( &g_matView, &vEyePt, &vLookatPt, &vUpVec );
    g_pd3dDevice->SetTransform( D3DTS_VIEW, &g_matView );

    ///   
	D3DXMatrixPerspectiveFovLH( &g_matProj, D3DX_PI/4, 1.0f, 1.0f, 1000.0f );
    g_pd3dDevice->SetTransform( D3DTS_PROJECTION, &g_matProj );

	///  ø  
    D3DXMatrixPerspectiveFovLH( &g_matProj, D3DX_PI/4, 1.0f, 1.0f, 200.0f );

	/// ī޶ ʱȭ
	g_pCamera->SetView( &vEyePt, &vLookatPt, &vUpVec );
}

/**-----------------------------------------------------------------------------
 * ؽó ʱȭ
 *------------------------------------------------------------------------------
 */
HRESULT InitTexture()
{
	// ̸ ؽó
	// D3DFMT_X8R8G8B8 D3DPOOL_MANAGED ֱؼ  Լ ߴ.
	if( FAILED( D3DXCreateTextureFromFileEx( g_pd3dDevice, BMP_HEIGHTMAP, 
								 D3DX_DEFAULT, D3DX_DEFAULT, 
								 D3DX_DEFAULT, 0, 
								 D3DFMT_X8R8G8B8, D3DPOOL_MANAGED, 
								 D3DX_DEFAULT, D3DX_DEFAULT, 0, 
								 NULL, NULL, &g_pTexHeight) ) )
		return E_FAIL;

	// 
	if( FAILED( D3DXCreateTextureFromFile( g_pd3dDevice, "tile2.tga", &g_pTexDiffuse) ) )
		return E_FAIL;

	return S_OK;
}

/**-----------------------------------------------------------------------------
 * ۸ ϰ  äִ´.
 * HeightMap ⼭ ʱȭѴ.
 *------------------------------------------------------------------------------
 */
HRESULT InitVB()
{
	D3DSURFACE_DESC		ddsd;
	D3DLOCKED_RECT		d3drc;

	g_pTexHeight->GetLevelDesc( 0, &ddsd );	// ؽó 
	g_cxHeight = ddsd.Width;				// ؽó ũ
	g_czHeight = ddsd.Height;				// ؽó ũ
	g_pLog->Log( "Texture Size:[%d,%d]", g_cxHeight, g_czHeight );
	g_pvHeightMap = new D3DXVECTOR3[g_cxHeight * g_czHeight];	// ̸ʹ迭 

	if( FAILED( g_pd3dDevice->CreateVertexBuffer( ddsd.Width*ddsd.Height*sizeof(CUSTOMVERTEX),
                                                  0, D3DFVF_CUSTOMVERTEX,
                                                  D3DPOOL_DEFAULT, &g_pVB, NULL ) ) )
    {
        return E_FAIL;
    }

	// ؽó ޸ !
	g_pTexHeight->LockRect( 0, &d3drc, NULL, D3DLOCK_READONLY );
    VOID* pVertices;
	//  !
    if( FAILED( g_pVB->Lock( 0, g_cxHeight*g_czHeight*sizeof(CUSTOMVERTEX), (void**)&pVertices, 0 ) ) )
        return E_FAIL;

	CUSTOMVERTEX	v;
	CUSTOMVERTEX*	pV = (CUSTOMVERTEX*)pVertices;
	for( DWORD z = 0 ; z < g_czHeight ; z++ )
	{
		for( DWORD x = 0 ; x < g_cxHeight ; x++ )
		{
			v.p.x = (float)x-g_cxHeight/2.0f;		//  xǥ(޽ø  )
			v.p.z = -((float)z-g_czHeight/2.0f);	//  zǥ(޽ø  ), z ;̹Ƿ - Ѵ.
			v.p.y = ((float)(*((LPDWORD)d3drc.pBits+x+z*(d3drc.Pitch/4))&0x000000ff))/10.0f;	/// DWORD̹Ƿ pitch/4
			v.n.x = v.p.x;
			v.n.y = v.p.y;
			v.n.z = v.p.z;
			D3DXVec3Normalize( &v.n, &v.n );
			v.t.x = (float)x / (g_cxHeight-1);
			v.t.y = (float)z / (g_czHeight-1);
			*pV++ = v;									// ۿ  
			g_pvHeightMap[z * g_cxHeight + x] = v.p;	// ̸ʿ  
		}
	}

    g_pVB->Unlock();
	g_pTexHeight->UnlockRect( 0 );

    return S_OK;
}

/**-----------------------------------------------------------------------------
 * ε  ʱȭ
 *------------------------------------------------------------------------------
 */
HRESULT InitIB()
{
    if( FAILED( g_pd3dDevice->CreateIndexBuffer( (g_cxHeight-1)*(g_czHeight-1)*2 * sizeof(MYINDEX), 0, D3DFMT_INDEX16, D3DPOOL_DEFAULT, &g_pIB, NULL ) ) )
    {
        return E_FAIL;
    }

	g_nTriangles = 0;

    return S_OK;
}

/**-----------------------------------------------------------------------------
 *  ʱȭ
 *------------------------------------------------------------------------------
 */
HRESULT InitGeometry()
{
	if( FAILED( InitTexture() ) ) return E_FAIL;
	if( FAILED( InitVB() ) ) return E_FAIL;
	if( FAILED( InitIB() ) ) return E_FAIL;

	InitMatrix();

	//  콺 ġ 
	POINT	pt;
	GetCursorPos( &pt );
	g_dwMouseX = pt.x;
	g_dwMouseY = pt.y;
	return S_OK;
}


/**-----------------------------------------------------------------------------
 * ʱȭ ü Ұ
 *------------------------------------------------------------------------------
 */
VOID Cleanup()
{
    if( g_pTexHeight != NULL )        
        g_pTexHeight->Release();

    if( g_pTexDiffuse!= NULL )        
        g_pTexDiffuse->Release();

    if( g_pIB != NULL )        
        g_pIB->Release();

    if( g_pVB != NULL )        
        g_pVB->Release();

    if( g_pd3dDevice != NULL ) 
        g_pd3dDevice->Release();

    if( g_pD3D != NULL )       
        g_pD3D->Release();

	delete[] g_pvHeightMap;
}


/**-----------------------------------------------------------------------------
 *  
 *------------------------------------------------------------------------------
 */
VOID SetupLights()
{
    /// (material)
    ///  ̽  ϳ   ִ.
    D3DMATERIAL9 mtrl;
    ZeroMemory( &mtrl, sizeof(D3DMATERIAL9) );
    mtrl.Diffuse.r = mtrl.Ambient.r = 1.0f;
    mtrl.Diffuse.g = mtrl.Ambient.g = 1.0f;
    mtrl.Diffuse.b = mtrl.Ambient.b = 1.0f;
    mtrl.Diffuse.a = mtrl.Ambient.a = 1.0f;
    g_pd3dDevice->SetMaterial( &mtrl );

    ///  
    D3DXVECTOR3 vecDir;									/// ⼺ (directional light)   
    D3DLIGHT9 light;									///  ü
    ZeroMemory( &light, sizeof(D3DLIGHT9) );			/// ü 0 .
    light.Type       = D3DLIGHT_DIRECTIONAL;			///  ( ,⼺ ,ƮƮ)
    light.Diffuse.r  = 1.0f;							///   
    light.Diffuse.g  = 1.0f;
    light.Diffuse.b  = 0.0f;
    vecDir = D3DXVECTOR3( 1, 1, 1 );					///  
    vecDir = D3DXVECTOR3(cosf(GetTickCount()/350.0f),	///  ȸ
                         1.0f,
                         sinf(GetTickCount()/350.0f) );
    D3DXVec3Normalize( (D3DXVECTOR3*)&light.Direction, &vecDir );	///   ͷ .
    light.Range       = 1000.0f;									///  ٴٸ ִ ִŸ
    g_pd3dDevice->SetLight( 0, &light );							/// ̽ 0  ġ
    g_pd3dDevice->LightEnable( 0, TRUE );							/// 0  Ҵ
    g_pd3dDevice->SetRenderState( D3DRS_LIGHTING, TRUE );			///  Ҵ

    g_pd3dDevice->SetRenderState( D3DRS_AMBIENT, 0x00909090 );		/// ȯ汤(ambient light)  
}

/**-----------------------------------------------------------------------------
 * Status 
 *------------------------------------------------------------------------------
 */
void LogStatus( void )
{
	g_pLog->Log( "Wireframe:%d", g_bWireframe );
	g_pLog->Log( "HideFrustum:%d", g_bHideFrustum );
	g_pLog->Log( "LockFrustum:%d", g_bLockFrustum );
}


/**-----------------------------------------------------------------------------
 * FPS(Frame Per Second)
 *------------------------------------------------------------------------------
 */
void LogFPS(void)
{
	static DWORD	nTick = 0;
	static DWORD	nFPS = 0;

	/// 1ʰ °?
	if( GetTickCount() - nTick > 1000 )
	{
		nTick = GetTickCount();
		/// FPS 
		g_pLog->Log("FPS:%d", nFPS );

		/// ī޶ ġ 
//		D3DXVECTOR3*	pv;
//		pv = g_pCamera->GetEye();
//		g_pLog->Log("EYE:[%f,%f,%f]",pv->x, pv->y, pv->z );
		nFPS = 0;
		LogStatus();	///  ⼭ (1ʿ ѹ)
		return;
	}
	nFPS++;
}


/**-----------------------------------------------------------------------------
 * Frustum Culling ؼ ׷ ε 
 *------------------------------------------------------------------------------
 */
HRESULT ProcessFrustumCull()
{
	DWORD		i[4];	// ӽ÷  ε 
	BOOL		b[4];	// ӽ÷  frustum culling
	MYINDEX		idx;
    MYINDEX*	pI;

    if( FAILED( g_pIB->Lock( 0, (g_cxHeight-1)*(g_czHeight-1)*2 * sizeof(MYINDEX), (void**)&pI, 0 ) ) )
        return E_FAIL;

	g_nTriangles = 0;

	for( DWORD z = 0 ; z < g_czHeight-1 ; z++ )
	{
		for( DWORD x = 0 ; x < g_cxHeight-1 ; x++ )
		{
			i[0] = (z*g_cxHeight+x);			//  
			i[1] = (z*g_cxHeight+x+1);			//  
			i[2] = ((z+1)*g_cxHeight+x);		//  ϴ
			i[3] = ((z+1)*g_cxHeight+x+1);		//  ϴ

			b[0] = g_pFrustum->IsIn( &g_pvHeightMap[ i[0] ] );	//   Frustumȿ ִ°?
			b[1] = g_pFrustum->IsIn( &g_pvHeightMap[ i[1] ] );	//   Frustumȿ ִ°?
			b[2] = g_pFrustum->IsIn( &g_pvHeightMap[ i[2] ] );	// ϴ  Frustumȿ ִ°?
			if( b[0] | b[1] | b[2] )	// ߿ ϳ frustumȿ  Ѵ.
			{
				idx._0 = i[0];
				idx._1 = i[1];
				idx._2 = i[2];
				*pI++ = idx;
				g_nTriangles++;			//  ﰢ  
			}

			b[2] = g_pFrustum->IsIn( &g_pvHeightMap[ i[2] ] );	// ϴ  Frustumȿ ִ°?
			b[1] = g_pFrustum->IsIn( &g_pvHeightMap[ i[1] ] );	//   Frustumȿ ִ°?
			b[3] = g_pFrustum->IsIn( &g_pvHeightMap[ i[3] ] );	// ϴ  Frustumȿ ִ°?
			if( b[2] | b[1] | b[3] )	// ߿ ϳ frustumȿ  Ѵ.
			{
				idx._0 = i[2];
				idx._1 = i[1];
				idx._2 = i[3];
				*pI++ = idx;
				g_nTriangles++;
			}
		}
	}
    g_pIB->Unlock();

    return S_OK;
}

/**-----------------------------------------------------------------------------
 * 콺 Է ó
 *------------------------------------------------------------------------------
 */
void ProcessMouse( void )
{
	POINT	pt;
	float	fDelta = 0.001f;	// 콺 ΰ,   Ŀ  δ.

	GetCursorPos( &pt );
	int dx = pt.x - g_dwMouseX;	// 콺 ȭ
	int dy = pt.y - g_dwMouseY;	// 콺 ȭ

	g_pCamera->RotateLocalX( dy * fDelta );	// 콺 Y ȸ 3D world  X ȸ
	g_pCamera->RotateLocalY( dx * fDelta );	// 콺 X ȸ 3D world  Y ȸ
	D3DXMATRIXA16*	pmatView = g_pCamera->GetViewMatrix();		// ī޶  ´.
	g_pd3dDevice->SetTransform( D3DTS_VIEW, pmatView );			// ī޶  


	// 콺  ߾ ʱȭ
//	SetCursor( NULL );	// 콺 Ÿ ʰ ʴ.
	RECT	rc;
	GetClientRect( g_hwnd, &rc );
	pt.x = (rc.right - rc.left) / 2;
	pt.y = (rc.bottom - rc.top) / 2;
	ClientToScreen( g_hwnd, &pt );
	SetCursorPos( pt.x, pt.y );
	g_dwMouseX = pt.x;
	g_dwMouseY = pt.y;
}

/**-----------------------------------------------------------------------------
 * Ű Է ó
 *------------------------------------------------------------------------------
 */
void ProcessKey( void )
{
	if( GetAsyncKeyState( 'A' ) ) g_pCamera->MoveLocalZ( 0.5f );	// ī޶ !
	if( GetAsyncKeyState( 'Z' ) ) g_pCamera->MoveLocalZ( -0.5f );	// ī޶ !
}

/**-----------------------------------------------------------------------------
 * Է ó
 *------------------------------------------------------------------------------
 */
void ProcessInputs( void )
{
	ProcessMouse();
	ProcessKey();
}

/**-----------------------------------------------------------------------------
 * ִϸ̼ 
 *------------------------------------------------------------------------------
 */
VOID Animate()
{
	D3DXMatrixIdentity( &g_matAni );

	SetupLights();
	ProcessInputs();

	D3DXMATRIXA16	m;
	D3DXMATRIXA16	*pView;
	pView = g_pCamera->GetViewMatrix();	// ī޶ Ŭκ  ´.
	m = *pView * g_matProj;				// Worldǥ ؼ View * Proj Ѵ.
	if( !g_bLockFrustum ) g_pFrustum->Make( &m );	// View*Projķ Frustum .
	ProcessFrustumCull();		// Frustum ,  ﰢ ε .

	LogFPS();
}


/**-----------------------------------------------------------------------------
 * ޽ ׸
 *------------------------------------------------------------------------------
 */
void DrawMesh( D3DXMATRIXA16* pMat )
{
	g_pd3dDevice->SetRenderState( D3DRS_FILLMODE, g_bWireframe ? D3DFILL_WIREFRAME : D3DFILL_SOLID );
    g_pd3dDevice->SetTransform( D3DTS_WORLD, pMat );
    g_pd3dDevice->SetStreamSource( 0, g_pVB, 0, sizeof(CUSTOMVERTEX) );
    g_pd3dDevice->SetFVF( D3DFVF_CUSTOMVERTEX );
    g_pd3dDevice->SetIndices( g_pIB );
	g_pd3dDevice->DrawIndexedPrimitive( D3DPT_TRIANGLELIST, 0, 0, g_cxHeight*g_czHeight, 0, g_nTriangles );
}

/**-----------------------------------------------------------------------------
 * ȭ ׸
 *------------------------------------------------------------------------------
 */
VOID Render()
{
    /// ĸۿ Z ʱȭ
    g_pd3dDevice->Clear( 0, NULL, D3DCLEAR_TARGET|D3DCLEAR_ZBUFFER, D3DCOLOR_XRGB(200,200,200), 1.0f, 0 );

	/// ִϸ̼ ļ
	Animate();
    ///  
    if( SUCCEEDED( g_pd3dDevice->BeginScene() ) )
    {
		g_pd3dDevice->SetTexture( 0, g_pTexDiffuse );							/// 0 ؽ  ؽ ()
		g_pd3dDevice->SetSamplerState( 0, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR );	/// 0 ؽó  Ȯ 
		g_pd3dDevice->SetTextureStageState( 0, D3DTSS_TEXCOORDINDEX, 0 );		/// 0 ؽó : 0 ؽó ε 

		g_pd3dDevice->SetTextureStageState( 0, D3DTSS_COLOROP,   D3DTOP_MODULATE);
		g_pd3dDevice->SetTextureStageState( 0, D3DTSS_COLORARG1, D3DTA_TEXTURE );
		g_pd3dDevice->SetTextureStageState( 0, D3DTSS_COLORARG2, D3DTA_DIFFUSE );

		DrawMesh( &g_matAni );
		if( !g_bHideFrustum ) g_pFrustum->Draw( g_pd3dDevice );

		///  
		g_pd3dDevice->EndScene();
    }

    /// ĸ۸ ̴ ȭ!
    g_pd3dDevice->Present( NULL, NULL, NULL, NULL );
}




/**-----------------------------------------------------------------------------
 *  ν
 *------------------------------------------------------------------------------
 */
LRESULT WINAPI MsgProc( HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam )
{
    switch( msg )
    {
        case WM_DESTROY :
            Cleanup();
            PostQuitMessage( 0 );
            return 0;
		case WM_KEYDOWN : 
			switch( wParam )
			{
				case VK_ESCAPE :
					PostMessage( hWnd, WM_DESTROY, 0, 0L );
					break;
				case '1' :
					g_bWireframe = !g_bWireframe;
					break;
				case '2' :
					g_bLockFrustum = !g_bLockFrustum;
					g_bHideFrustum = !g_bLockFrustum;
					break;
			}
			break;
    }

    return DefWindowProc( hWnd, msg, wParam, lParam );
}




/**-----------------------------------------------------------------------------
 * α׷ 
 *------------------------------------------------------------------------------
 */
INT WINAPI WinMain( HINSTANCE hInst, HINSTANCE, LPSTR, INT )
{
    ///  Ŭ 
    WNDCLASSEX wc = { sizeof(WNDCLASSEX), CS_CLASSDC, MsgProc, 0L, 0L,
                      GetModuleHandle(NULL), NULL, NULL, NULL, NULL,
                      "BasicFrame", NULL };
    RegisterClassEx( &wc );

    ///  
    HWND hWnd = CreateWindow( "BasicFrame", WINDOW_TITLE,
                              WS_OVERLAPPEDWINDOW, 100, 100, WINDOW_W, WINDOW_H,
                              GetDesktopWindow(), NULL, wc.hInstance, NULL );

	g_hwnd = hWnd;

	g_pLog = new ZFLog( ZF_LOG_TARGET_CONSOLE | ZF_LOG_TARGET_WINDOW );
	g_pCamera = new ZCamera;
	g_pFrustum = new ZFrustum;

    /// Direct3D ʱȭ
    if( SUCCEEDED( InitD3D( hWnd ) ) )
    {
        if( SUCCEEDED( InitGeometry() ) )
        {

        	///  
			ShowWindow( hWnd, SW_SHOWDEFAULT );
			UpdateWindow( hWnd );

        	/// ޽ 
			MSG msg;
			ZeroMemory( &msg, sizeof(msg) );
			while( msg.message!=WM_QUIT )
			{
            	/// ޽ť ޽  ޽ ó
				if( PeekMessage( &msg, NULL, 0U, 0U, PM_REMOVE ) )
				{
					TranslateMessage( &msg );
					DispatchMessage( &msg );
				}
				else
				/// ó ޽  Render()Լ ȣ
					Render();
			}
		}
    }

	/// ϵ Ŭ Ұ
	delete g_pFrustum;
	delete g_pLog;
	delete g_pCamera;
    UnregisterClass( "D3D Tutorial", wc.hInstance );
    return 0;
}
